/* $Id: dyn_array.c,v 6.4 1999/08/30 20:22:53 cher Exp $ */

/* Copyright (C) 1997 SDL Research Lab, Moscow */

/*
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 */

/**
 * FILE:        dyn_array.c
 * AUTHOR:      dmjr
 * SYSTEM:      Template Classes
 * DESCRIPTION: The dynamic array class library
 */

/* -----------------------Include-files----------------------- */

static char const cvsid[] =
"$Id: dyn_array.c,v 6.4 1999/08/30 20:22:53 cher Exp $";

#include "lconfig.h"
#include <p_config.h>

#include CONF_DEPFILE("reuse_version.h")

#include "reuse/Defs.h"
#include "reuse/dyn_array.h"
#include "reuse/xalloc.h"

#include <string.h>

/* ---------------------Macro-definitions--------------------- */

/* Initial non-empty array size */

#define START_SIZE 4

/* ---------------------Exported-variables-------------------- */

/* --------------------Function-definitions------------------- */

/**
 * NAME:      da_Add
 * PURPOSE:   add an element to the end of an array
 * SYNOPSIS:  tpDynamicArray array -- the array
 *            size_t eltSize -- element size
 * RETURNS:   void * -- pointer to added element
 */

void *
da_Add (tpDynamicArray array, size_t eltSize)
{
  da_Extend (array, array->size + 1, eltSize);
  return ((char *)array->v) + eltSize * (array->size - 1);
}

/**
 * NAME:      da_Extend
 * PURPOSE:   extend an array to given size
 * SYNOPSIS:  tpDynamicArray array -- the array
 *            newSize -- the new size
 *            size_t eltSize -- element size
 * RETURNS:   void
 * NOTE:      if the array already contains more than size elements
 *                function does nothing
 */

void
da_Extend (tpDynamicArray array, size_t newSize, size_t eltSize)
{
  void *newVec;

  if (newSize > array->alloc) {
    if (array->alloc == 0)
      array->alloc = START_SIZE;
    while (newSize > array->alloc)
      array->alloc *= 2;
    newVec = xrealloc (array->v, array->alloc * eltSize);
    array->v = newVec;
  }
  array->size = newSize;
}

/**
 * NAME:      da_Copy
 * PURPOSE:   copy one array to another
 * SYNOPSIS:  tpDynamicArray dst -- the source array
 *            tpDynamicArray src -- the destanation array
 *            size_t eltSize -- element size
 * RETURNS:   void
 */

void
da_Copy (tpDynamicArray dst, tpDynamicArray src, size_t eltSize)
{
  da_Extend (dst, src->size, eltSize);
  memcpy (dst->v, src->v, src->size * eltSize);
}

/**
 * NAME:      da_Freeze
 * PURPOSE:   release the most possible memory amount used by array
 * SYNOPSIS:  tpDynamicArray array -- the array
 *            size_t eltSize -- element size
 * RETURNS:   void
 */

void
da_Freeze (tpDynamicArray array, size_t eltSize)
{
  size_t newAlloc;

  if (array->size == 0) {

    array->alloc = 0;
    free (array->v);
    array->v = NULL;

  } else {

    newAlloc = array->alloc;
    for (;;) {
      newAlloc /= 2;
      if (newAlloc < array->size ||
          newAlloc < START_SIZE) {
        newAlloc *= 2;
        break;
      }
    }
    if (newAlloc != array->alloc) {
      array->alloc = newAlloc;
      array->v = realloc (array->v, array->alloc * eltSize);
    }
  }
}

/**
 * NAME:      da_AppendMem
 * PURPOSE:   append memory vector to an array
 * SYNOPSIS:  tpDynamicArray dst -- destanation array
 *            void *vec -- pointer to vector
 *            size_t vecSize -- size of the vector
 *            size_t eltSize -- element size
 * RETURNS:   void
 */

void
da_AppendMem (tpDynamicArray dst, void *vec, size_t vecSize,
              size_t eltSize)
{
  size_t oldSize = dst->size;

  da_Extend (dst, dst->size + vecSize, eltSize);
  memcpy (dst->v + oldSize * eltSize, vec, vecSize * eltSize);
}

/* === Do not add stuff after this line === */
