/* $Id: HeapMem.c,v 6.8 2001/06/23 07:03:05 cher Exp $ */

/* Copyright (C) 1995-2001 Alexander Chernov <cher@ispras.ru> */
/* Created: <95/12/03 15:51:29 cher> */

/*
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 */

static char const rcsid[]=
"$Id: HeapMem.c,v 6.8 2001/06/23 07:03:05 cher Exp $";

#include "lconfig.h"
#include <p_config.h>
#include <p_integral.h>

#include CONF_DEPFILE("reuse_version.h")

#include "reuse/ReuseDefs.h"
#include "reuse/xalloc.h"
#include "reuse/HeapMem.h"

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>

#define MAGIC      0x220176cc
#define CHECK_BYTE 0xcd
#define CHECK_WORD 0xcdcdcdcd
#define ALIGN(x)   (((x) + 3) & ~3)

struct HEAP_NODE;

typedef struct HEAP_HEAD
{
  rulong_t Magic;
  rulong_t Size;
  
#if defined(KEEP_LIST)
  struct HEAP_NODE *Next;
  struct HEAP_NODE *Prev;
#endif

#ifdef KEEP_CALLER
  char   *File;
  line_t Line;
#endif
  
  rulong_t LowDummy;
} HEAP_HEAD;

typedef struct HEAP_NODE
{
  HEAP_HEAD Head;
  rulong_t Data[1];
} HEAP_NODE;

#if defined(KEEP_LIST)
static HEAP_NODE * heap_list;
#endif

#ifdef COUNT_MEM
static rulong_t allocated_size;
static rulong_t freed_size;
static rulong_t internal_size;
#endif

static Bool initialized;

static void do_init(void);
static void heap_exhausted(void);

  static void
check_node (HEAP_NODE *node)
{
  if (node->Head.Magic != MAGIC)
    {
      fputs ("Not owned or already freed heap block!\n",stderr);
      abort ();
    }

  if (node->Head.LowDummy != CHECK_WORD)
    {
      fputs ("Access before low boudary of heap memory block!\n",stderr);
      abort ();
    }

  if(((ruchar_t*) node->Data)[node->Head.Size] != CHECK_BYTE)
    {
      fputs ("Access after high boundary of heap memory block!\n",stderr);
      abort ();
    }
}

  void*
_checked_malloc(size_t size)
{
  HEAP_NODE *heap_ptr;
  size_t aligned_size = ALIGN(size);

  if (!initialized) do_init();
  assert(size != 0);
  
  heap_ptr = (HEAP_NODE*) malloc (sizeof (HEAP_NODE)+aligned_size);
  if (heap_ptr == NULL)
    heap_exhausted ();

  heap_ptr->Head.Magic = MAGIC;
  heap_ptr->Head.Size = size;

#ifdef COUNT_MEM
  allocated_size += size;
  internal_size  += sizeof(HEAP_NODE) + aligned_size;
#endif

#ifdef KEEP_CALLER
  heap_ptr->Head.File = CALLER_FILE;
  heap_ptr->Head.Line = CALLER_LINE;
#endif
  
#if defined(KEEP_LIST)
  if (heap_list != NULL)
    heap_list->Head.Prev = heap_ptr;
  heap_ptr->Head.Next = heap_list;
  heap_ptr->Head.Prev = NULL;
  heap_list = heap_ptr;
#endif

  memset ((char*) &heap_ptr->Head.LowDummy,
          CHECK_BYTE,
         aligned_size + 2*sizeof (rulong_t));
  
  return (void*) &heap_ptr->Data;
}

  void*
_checked_calloc(size_t nitems, size_t elem_size)
{
  HEAP_NODE *heap_ptr;
  size_t size = nitems * elem_size;
  size_t aligned_size = ALIGN(size);

  if (!initialized) do_init();
  assert(size != 0);
  
  heap_ptr = (HEAP_NODE*) malloc (sizeof (HEAP_NODE)+aligned_size);
  if (heap_ptr == NULL)
    heap_exhausted ();

  heap_ptr->Head.Magic = MAGIC;
  heap_ptr->Head.Size = size;

#ifdef COUNT_MEM
  allocated_size += size;
  internal_size  += aligned_size + sizeof(HEAP_NODE);
#endif

#ifdef KEEP_CALLER
  heap_ptr->Head.File = CALLER_FILE;
  heap_ptr->Head.Line = CALLER_LINE;
#endif
  
#if defined(KEEP_LIST)
  if (heap_list != NULL)
    heap_list->Head.Prev = heap_ptr;
  heap_ptr->Head.Next = heap_list;
  heap_ptr->Head.Prev = NULL;
  heap_list = heap_ptr;
#endif
  
  memset ((char*) &heap_ptr->Head.LowDummy,
         CHECK_BYTE,
         aligned_size + 2*sizeof (rulong_t));
  memset ((char*) &heap_ptr->Data, 0, size);
  
  return (void*) &heap_ptr->Data;
}

  void
_checked_free(void *ptr)
{
  HEAP_NODE * heap_ptr;
  
  if (!initialized) do_init();
  assert (ptr != NULL);

  heap_ptr = (HEAP_NODE*) ((char*) ptr - sizeof(HEAP_HEAD));
  check_node (heap_ptr);

#ifdef COUNT_MEM
  freed_size += heap_ptr->Head.Size;
#endif
  
#if defined(KEEP_LIST)
  if(heap_ptr->Head.Prev == NULL)
    {
      heap_list = heap_ptr->Head.Next;
      if(heap_list != NULL)
        heap_list->Head.Prev = NULL;
    }
  else
    {
      HEAP_NODE *prev_node = heap_ptr->Head.Prev;
      
      prev_node->Head.Next = heap_ptr->Head.Next;
      if(heap_ptr->Head.Next != NULL)
        prev_node->Head.Next->Head.Prev = prev_node;
    }
#endif

  memset ((char*) heap_ptr,
          CHECK_BYTE,
          ALIGN (heap_ptr->Head.Size) + sizeof(HEAP_NODE));

  free (heap_ptr);
}

  void*
_checked_realloc(void *ptr, size_t size)
{
  void *new_ptr;
  
  if (!initialized) do_init();

  if(size == 0)
    {
      _checked_free(ptr);
      return NULL;
    }

  if(ptr == NULL)
    {
      return _checked_malloc(size);
    }

  new_ptr = _checked_malloc(size);
  
  memcpy (new_ptr, ptr, size);
  _checked_free(ptr);
  return new_ptr;
}

  void*
_checked_expand(void *ptr, size_t size)
{
  HEAP_NODE * heap_ptr;
  void *new_ptr;
  
  if (!initialized) do_init();
  assert (ptr != NULL && size != 0);

  heap_ptr = (HEAP_NODE*) ((char*) ptr - sizeof(HEAP_HEAD));
  check_node (heap_ptr);

  if (heap_ptr->Head.Size >= size)
    {
      fputs ("Invalid expand size!\n", stderr);
      abort ();
    }

  new_ptr = _checked_malloc(size);
  
  memcpy (new_ptr, ptr, size);
  _checked_free(ptr);
  return new_ptr;
}

  void*
_checked_shrink(void *ptr, size_t size)
{
  HEAP_NODE * heap_ptr;
  void *new_ptr;
  
  assert (ptr != NULL);
  if (!initialized) do_init();

  heap_ptr = (HEAP_NODE*) ((char*) ptr - sizeof(HEAP_HEAD));
  check_node (heap_ptr);

  if (heap_ptr->Head.Size >= size)
    {
      fputs ("Invalid shrink size!\n", stderr);
      abort ();
    }  
  
  new_ptr = _checked_malloc(size);

  memcpy (new_ptr, ptr, size);
  _checked_free(ptr);
  return new_ptr;
}

  char*
_checked_strdup(char *str)
{
  int len;
  char *ret_ptr;

  assert(str != NULL);
  if (!initialized) do_init();

  len = strlen(str);
  ret_ptr = (char*) _checked_malloc(len + 1);
  assert(ret_ptr != NULL);
  strcpy(ret_ptr, str);
  return ret_ptr;
}

/* NULL checking functions */

  void*
_NULL_checked_malloc(size_t size)
{
  return xmalloc(size);
}

  void*
_NULL_checked_calloc(size_t nitems, size_t size)
{
  return xcalloc(nitems, size);
}

  void*
_NULL_checked_realloc(void *ptr, size_t new_size)
{
  return xrealloc(ptr, new_size);
}

  char*
_NULL_checked_strdup(char *str)
{
  return xstrdup(str);
}

/* Initialization and statistics functions */

  void
chInitModule(void)
{
  if (!initialized) do_init();
}

  static void
do_init(void)
{
  initialized = True;
#ifdef KEEP_LIST
  heap_list = NULL;
#endif
      
#ifdef COUNT_MEM
  allocated_size = 0;
  internal_size  = 0;
  freed_size     = 0;
#endif
}

  void
chStatistics(FILE *out)
{
#ifdef REUSE_DEBUG
#ifdef KEEP_LIST
  HEAP_NODE *list_ptr;
#endif /* KEEP_LIST */

  fprintf(out, "Heap memory usage statistics:\n");

#ifdef COUNT_MEM
  fprintf(out, "Total size of allocation requests: %lu\n", allocated_size);
  fprintf(out, "Total size of freed memory:        %lu\n", freed_size);
  fprintf(out, "Memory leak:                       %lu\n",
          allocated_size - freed_size);
  fprintf(out, "Size of allocated memory:          %lu\n", internal_size);
#endif /* COUNT_MEM */

#ifdef KEEP_LIST
#ifdef KEEP_CALLER
  fprintf(out, "List of functions where leak memory was allocated:\n");
#endif /* KEEP_CALLER */

  for (list_ptr = heap_list;
       list_ptr != NULL;
       list_ptr = list_ptr->Head.Next)
    {
      check_node(list_ptr);
#ifdef KEEP_CALLER
      fprintf(out, "%s:%d: size = %lu\n",
              list_ptr->Head.File,
              list_ptr->Head.Line,
              list_ptr->Head.Size);
#endif /* KEEP_CALLER */
    }

#endif /* KEEP_LIST */
#endif /* REUSE_DEBUG */
}

  void
chCloseModule(void)
{
  initialized = False;
}

  static void
heap_exhausted(void)
{
  fputs("Heap memory exhausted!\n",stderr);
  abort();
}

/**
 * Local variables:
 *  compile-command: "make -C .."
 *  tags-file-name: "../TAGS"
 *  c-font-lock-extra-types: ("FILE" "\\sw+_t" "HEAP_NODE")
 * End:
 */
