/* $Id: nstream_io.c,v 6.6 2001/06/23 07:03:05 cher Exp $ */

/* Copyright (C) 1998-2001 Alexander Chernov <cher@ispras.ru> */

/*
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 */

static char const cvsid[] =
"$Id: nstream_io.c,v 6.6 2001/06/23 07:03:05 cher Exp $";

#include "lconfig.h"
#include <p_config.h>

#include CONF_DEPFILES("reuse_version.h")

#include "reuse/nstream_io.h"
#include "reuse/nstream_private.h"
#include "reuse/xalloc.h"
#include "reuse/osdeps.h"

#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <errno.h>
#include <netinet/in.h>
#include <stdarg.h>

/* extra error codes */
#define E_STRMFIRST     2048

#define ENOSTREAM       2048
#define ENOREAD         2049
#define ENOWRITE        2050
#define ENOVIRTUAL      2051
#define EFNOTAVAIL      2052
#define ETOOLATE        2053

/* unidirected stream */
struct stream
{
        /* flags specified upon creation */
        int     flags;

        /* buffering */
        char    *buf;           /* buffer pointer */
        char    *ptr;           /* current buffer pointer */
        char    *eptr;          /* pointer to the last byte in buffer */
        int     size;           /* size of the buffer */
        int     iseof;          /* end of file indicator */

        /* virtual function table */
        struct stream_vtable *vt;

        /* extra low-level data */
        void    *data;
};

/* Generic stream structure (for all kinds of communications) */
struct tStream
{
        struct stream   *r;     /* read stream structure */
        struct stream   *w;     /* write stream structure */
        int             status; /* status of the last operation */
        struct stream   *orig;  /* originator of the status */
};

struct stream_vtable stream_file_vtable;

static struct stream_vtable *vtables[] =
{
        0,
        &stream_file_vtable,

        0
};

#define DEFAULT_BUF_SIZE 1024

enum
{
        STC_BIDIR       = 0x00000001,   /* bidirectional input/output */
        STC_NOBUF       = 0x00000002,   /* not buffered */
        STC_RFLUSH      = 0x00000004,   /* flush buffer on read */
};

        struct tStream*
nstrm_Create(int rf, int rm, int wf, int wm)
{
        struct tStream          *p;
        struct stream_vtable    *read_table = 0;
        struct stream_vtable    *write_table = 0;
        int                     i;

        if (!rm && !wm) return 0;

        for (i = 0; vtables[i] != 0; i++) {
                if (i == rm) read_table = vtables[rm];
                if (i == wm) read_table = vtables[wm];
        }
        if (rm && !read_table) return 0;
        if (wm && !write_table) return 0;

        if ((wf & STC_BIDIR)) {
                write_table = read_table;
        }

        if (read_table && !read_table->create) return 0;
        if (write_table && !write_table->create) return 0;

        p = xmalloc(sizeof (struct tStream));
        p->status = 0;
        p->orig   = 0;
        p->w      = 0;
        p->r      = 0;

        if (read_table) {
                p->r = xmalloc(sizeof(struct stream));
                p->r->flags = rf;
                p->r->buf = 0;
                p->r->ptr = 0;
                p->r->iseof = 0;
                if ((rf & STC_NOBUF)) {
                        p->r->size = 0;
                } else {
                        p->r->size = DEFAULT_BUF_SIZE;
                }
                p->r->vt = read_table;
                p->r->data = 0;
                p->r->data = read_table->create(p->r->data);
        }
        if (write_table) {
                p->w = xmalloc(sizeof(struct stream));
                p->w->flags = wf;
                p->w->buf = 0;
                p->w->ptr = 0;
                p->w->iseof = 0;
                if ((wf & STC_NOBUF)) {
                        p->w->size = 0;
                } else {
                        p->w->size = DEFAULT_BUF_SIZE;
                }
                p->w->vt = write_table;
                if ((wf & STC_BIDIR)) {
                        p->w->data = p->r->data;
                } else {
                        p->w->data = 0;
                }
                p->w->data = write_table->create(p->w->data);
        }

        return p;
}

        int
nstrm_Flush(struct tStream *s)
{
        int r;

        if (!s) return -ENOSTREAM;
        if (s->status) return -s->status;
        if (!s->w) {
                s->status = ENOWRITE;
                s->orig = 0;
                return -ENOWRITE;
        }

        if (!s->w->size) return 0;
        if (!s->w->buf) return 0;
        if (s->w->buf == s->w->ptr) return 0;

        if (!s->w->vt->write) {
                s->status = ENOVIRTUAL;
                s->orig = s->w;
                return -ENOVIRTUAL;
        }
        
        if (s->w->iseof) return 0;
        r = s->w->vt->write(s->w->data, s->w->buf, s->w->ptr - s->w->buf);
        if (r < 0) {
                s->status = -r;
                s->orig = s->w;
                return r;
        }
        if (r == 0) s->w->iseof = 1;
        s->w->ptr = s->w->buf;
        return r;
}

        int
nstrm_PutRawBytes(struct tStream *s, void *data, int n)
{
        int r;
        int ws;
        int wb = 0;
        unsigned short *cdata = (unsigned short *) data;

        if (!s) return -ENOSTREAM;
        if (s->status) return -s->status;
        if (!s->w) {
                s->status = ENOWRITE;
                s->orig = 0;
                return -ENOWRITE;
        }

        if (n <= 0 || !data) {
                s->status = EINVAL;
                s->orig = s->w;
                return -EINVAL;
        }

        if (s->w->iseof) return 0;

        /* write directly, if output is unbuffered */
        if (!s->w->size) {
                if (!s->w->vt->write) {
                        s->status = ENOVIRTUAL;
                        s->orig = s->w;
                        return -ENOVIRTUAL;
                }

                r = s->w->vt->write(s->w->data, cdata, n);
                if (r >= 0) {
                        if (r == 0) s->w->iseof = 1;
                        return r;
                }
                s->status = r;
                s->orig = s->w;
                return -r;
        }

        /* create buffer, if such not exist */
        if (!s->w->buf) {
                s->w->buf = xmalloc(s->w->size);
                s->w->ptr = s->w->buf;
        }

        while (n > 0) {
                /* if buffer is not empty, append to the buffer */
                if (s->w->ptr != s->w->buf) {
                        ws = s->w->buf + s->w->size - s->w->ptr;
                        if (ws > n) ws = n;
                        memcpy(s->w->ptr, cdata, ws);
                        s->w->ptr += ws;
                        cdata += ws;
                        n -= ws;
                        wb += ws;
                } else if (n >= s->w->size) {
                /* buffer is empty and block is larger than buffer */
                        ws = n / s->w->size;
                        ws *= s->w->size;
                        if (!s->w->vt->write) {
                                s->status = ENOVIRTUAL;
                                s->orig = s->w;
                                return -ENOVIRTUAL;
                        }
                        r = s->w->vt->write(s->w->data, cdata, ws);
                        if (r < 0) {
                                s->status = -r;
                                s->orig = s->w;
                                return r;
                        }
                        if (r == 0) {
                                s->w->iseof = 1;
                                return 0;
                        }
                        wb += r;
                        n -= r;
                } else {
                /* buffer is empty, block fits to the buffer */
                        memcpy(s->w->ptr, cdata, n);
                        s->w->ptr += n;
                        wb += n;
                        n = 0;
                }

                /* check, that we should flush buffer */
                if (s->w->buf + s->w->size == s->w->ptr) {
                        if ((r = nstrm_Flush(s)) < 0)
                                return r;
                }
        }
        return wb;
}

        int
nstrm_PutChar(struct tStream *s, int c)
{
        unsigned char cc = (unsigned char) c;

        return nstrm_PutRawBytes(s, &cc, 1);
}

        int
nstrm_PutShort(struct tStream *s, int c)
{
#if CONF_ENDIAN == 1
        /* big endian machine */
        short b = (short) c;
        return nstrm_PutRawBytes(s, &b, 2);
#elif CONF_ENDIAN == 2
        /* little endian machine */
        unsigned char b[2], *pc = (unsigned char*) &c;
        b[0] = pc[1];
        b[1] = pc[0];
        return nstrm_PutRawBytes(s, &b, 2);
#else
        /* use default conversion routine */
        short b = htons(c);
        return nstrm_PutRawBytes(s, &b, 2);
#endif /* CONF_ENDIAN */
}

        int
nstrm_PutInt(struct tStream *s, int c)
{
#if CONF_ENDIAN == 1
        /* big endian machine */
        return nstrm_PutRawBytes(s, &c, 4);
#elif CONF_ENDIAN == 2
        /* little endian machine */
        unsigned char d[4], *pd = d + 4, *pc = (unsigned char*) &c;
        while (pd > d) *--pd = *pc++;
        return nstrm_PutRawBytes(s, &d, 4);
#else
        long b = htonl(c);
        return nstrm_PutRawBytes(s, &b, 4);
#endif /* CONF_ENDIAN */
}

        int
nstrm_PutNetBytes(struct tStream *s, void *b, int n)
{
        if (!s) return -ENOSTREAM;
        if (s->status) return -s->status;
        if (!s->w) {
                s->status = ENOWRITE;
                s->orig = s->w;
                return -ENOWRITE;
        }
        if (n != 1 && n != 2 && n != 4 && n != 8) {
                s->status = EINVAL;
                s->orig = s->w;
                return -EINVAL;
        }
#if CONF_ENDIAN == 1
        /* big endian */
        return nstrm_PutRawBytes(s, b, n);
#elif CONF_ENDIAN == 2
        /* little endian */
        pd = d + n;
        pb = (unsigned char*) b;
        while (pd > d) *--pd = *pb++;
#else
        s->status = EFNOTAVAIL;
        s->orig = s->w;
        return -EFNOTAVAIL;
#endif
}

#if CONF_LONGLONG == 1
        int
nstrm_PutLong(s, c)
        struct tStream *s;
        long long c;
{
        return nstrm_PutNetBytes(s, &c, 8);
}
#else /* CONF_LONGLONG */
        int
nstrm_PutLong(s)
        struct tStream *s;
{
        if (!s) return -ENOSTREAM;
        if (s->status) return -s->status;
        if (!s->w) {
                s->status = ENOWRITE;
                s->orig = 0;
                return -ENOWRITE;
        }

        s->status = EFNOTAVAIL;
        s->orig = s->w;
        return -EFNOTAVAIL;
}
#endif /* CONF_LONGLONG */

        int
nstrm_Write(struct tStream *s, void *p, int n)
{
        return nstrm_PutRawBytes(s, p, n);
}

        int
nstrm_Puts(struct tStream *s, char *p)
{
        int len = p?strlen(p):0;
        return nstrm_PutRawBytes(s, p, len);
}

#define PRINTF_BUF_SIZE 16384
static char *printf_buf;
static char const errmsg[] = "Fatal buffer overrun in nstrm_vPrintf\n";
        int
nstrm_vPrintf(struct tStream *s, char *format, va_list args)
{
        int     len;

        if (!printf_buf) printf_buf = xmalloc(PRINTF_BUF_SIZE);
        len = sprintf(printf_buf, format, args);
        if (len >= PRINTF_BUF_SIZE) {
                /* buffer overrun!!! */
                /*write(2, errmsg, sizeof(errmsg) - 1);*/
                abort(); /* FIXME */
        }
        return nstrm_PutRawBytes(s, printf_buf, len);
}

        int
nstrm_Printf(struct tStream *s, char *format, ...)
{
        int     c;
        va_list args;

        va_start(args, format);
        c = nstrm_vPrintf(s, format, args);
        va_end(args);
        return c;
}

        int
nstrm_GetRawBytes(struct tStream *s, void *data, int n)
{
        int             r;
        int             rb = 0;
        unsigned char   *cdata = (unsigned char *) data;

        if (!s) return -ENOSTREAM;
        if (s->status) return -s->status;
        if (!s->r) {
                s->status = ENOREAD;
                s->orig = 0;
                return -ENOREAD;
        }

        if (n <= 0 || !data) {
                s->status = EINVAL;
                s->orig = s->r;
                return -EINVAL;
        }

        if (s->r->iseof) return 0;

        if (!s->r->size) {
                /* read directly (input is unbuffered) */
                if (!s->r->vt->read) {
                        s->status = ENOVIRTUAL;
                        s->orig = s->r;
                        return -ENOVIRTUAL;
                }

                r = s->r->vt->read(s->r->data, cdata, n);
                if (r >= 0) {
                        if (r == 0) s->r->iseof = 1;
                        return r;
                }
                s->status = -r;
                s->orig = s->r;
                return r;
        }

        /* read through buffer */
        /* Create buffer, if it does not exist */
        if (!s->r->buf) {
                s->r->buf = xmalloc(s->r->size);
                s->r->ptr = s->r->buf + s->r->size;
        }

        /* if buffer is not empty, read from buffer */
        if (s->r->ptr < s->r->buf + s->r->size) {
                rb = s->r->buf + s->r->size - s->r->ptr;
                if (n < rb) rb = n;
                memcpy(cdata, s->r->ptr, rb);
                s->r->ptr += rb;
                n -= rb;
                cdata += rb;
        }
        return 0;
}

#define ERRBUFSIZE 1024
static char *errbuf = 0;

        char*
nstrm_StrError(int err)
{
        if (err == 0) return "No error";
        if (err < E_STRMFIRST) return os_GetErrorString(err);
        switch (err) {
        case ENOSTREAM:
                return "Stream pointer is NULL";
        case ENOWRITE:
                return "Stream is not opened for write";
        case ENOREAD:
                return "Stream is not opened for read";
        case ENOVIRTUAL:
                return "Virtual function is not implemented";
        case EFNOTAVAIL:
                return "Not available on this architecture";
        case ETOOLATE:
                return "Too late";
        default:
                if (!errbuf) errbuf = xmalloc(ERRBUFSIZE);
                sprintf(errbuf, "Unknown stream error (%d)", err);
                return errbuf;
        }
}

        char*
nstrm_Message(struct tStream *s)
{
        return nstrm_StrError(s?s->status:-ENOSTREAM);  
}

        int
nstrm_Status(struct tStream *s)
{
        return s?s->status:-ENOSTREAM;
}

        int
nstrm_Recover(struct tStream *s)
{
        int r;

        if (!s) return -ENOSTREAM;
        s->status = 0;
        if (!s->orig) return 0;
        if (!s->orig->vt->recover) {
                s->orig = 0;
                return 0;
        }

        r = s->orig->vt->recover(s->orig->data);
        if (r < 0) {
                s->status = -r;
                return r;
        }
        return 0;
}

/* === Do not add stuff after this line === */

