/* -*- mode:c -*- */
/* $Id: parse_printf.c,v 1.1 2002/06/08 08:27:45 cher Exp $ */

/* Copyright (C) 2002 Alexander Chernov <cher@ispras.ru> */

/*
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 */

#include "lconfig.h"
#include <p_config.h>
#include <p_integral.h>

#include "reuse/format_io.h"

/**
 * Parses one printf specification.
 * Returns the conversion specifier (one of 'c', 'd', 'e', 'E',
 *   'f', 'g', 'G', 'i', 'o', 'p', 's', 'u', 'x', 'X', 'a', 'A'),
 *   or -1 in case of unrecognized conversion specifier
 * Parameters are as follows:
 *   str - string to extract the first parameter
 *   eptr - pointer to store the next-after-last character posision
 *   past_num - number of asterisks in specification (0 - 2)
 *   parg_size - argument size specifier ('H', 'h', 'l', 'L', 'q',
 *               'z', 'j')
 *   pw_flag - how width was specified: 0 - none, 1 - explicitly,
 *             2 - via '*'
 *   pw - width
 *   pp_flag - how precision was specified: 0 - none, 1 - explicitly,
 *             2 - via '*'
 *   pp - precision
 */
int
reuse_parse_printf(char const *str,
                   char **eptr,
                   int *past_num,
                   int *parg_size,
                   int *pw_flag,
                   int *pw,
                   int *pp_flag,
                   int *pp)
{
  int ret_val = -1;
  int ast_n = 0;
  int arg_size = 0;
  int p_flag = 0, p = 0, w_flag = 0, w = 0;
  int state = 0;

  /* 0 - start (reading flags) */
  /* 1 - reading width (number) */
  /* 2 - after width asterisk */
  /* 3 - after reading dot */
  /* 4 - reading prec (number) */
  /* 5 - after prec asterisk */

  if (*str != '%') goto _exit;
  str++;

  while (1) {
    switch (*str) {
    case '*':
      /* need to know how many asterisks was */
      if (ast_n == 2) {
        ast_n = 0;
        goto _exit;
      }
      ast_n++;
      if (!state) {
        w_flag = 2;
        state = 2;
      } else if (state == 3) {
        p_flag = 2;
        state = 5;
      } else {
        goto _exit;
      }
      break;
    case ' ': case '+': case '-': case '#':
      /* flags */
      if (state) goto _exit;
      break;
    case '.':
      if (state >= 3) goto _exit;
      state = 3;
      break;

      /* it is either a flag, or a part of width or precision */
    case '0':
      if (!state) break;
      /* FALLTHROUGH */
    case '1': case '2': case '3': case '4':
    case '5': case '6': case '7': case '8': case '9':
      /* field width */
      if (state == 0 || state == 1) {
        state = 1;
        w_flag = 1;
        if (w > 100000) goto _exit;
        w = w * 10 + (*str - '0');
      } else if (state == 3 || state == 4) {
        state = 4;
        p_flag = 1;
        if (p > 100000) goto _exit;
        p = p * 10 + (*str - '0');
      } else {
        goto _exit;
      }
      break;
    case 'h':
      if (arg_size == 'h') {
        arg_size = 'H';
      } else if (arg_size) {
        goto _exit;
      } else {
        arg_size = *str;
      }
      break;
    case 'l':
      if (arg_size == 'l') {
        arg_size = 'L';
      } else if (arg_size) {
        goto _exit;
      } else {
        arg_size = *str;
      }
      break;
    case 'L': case 'z': case 'j': case 'q':
      /* these argument size cannot repeat */
      if (arg_size) goto _exit;
      arg_size = *str;
      break;
      /* conversion modes */
    case 'c': case 'd': case 'e': case 'E':
    case 'f': case 'g': case 'G': case 'i':
    case 'o': case 'p': case 's':
    case 'u': case 'x': case 'X': case 'a':
    case 'A':
      ret_val = *str++;
      goto _exit;
    case 'n':
      /* this is explicitly not supported */
      goto _exit;
    case 0:
    default:
      /* invalid format string */
      goto _exit;
    }
    str++;
  }

 _exit:
  if (past_num) *past_num = ast_n;
  if (pp_flag) *pp_flag = p_flag;
  if (pp) *pp = p;
  if (pw_flag) *pw_flag = w_flag;
  if (pw) *pw = w;
  if (parg_size) *parg_size = arg_size;
  if (eptr) *eptr = (char*) str;
  return ret_val;
}

/*
 * Local variables:
 *  compile-command: "make -C .."
 * End:
 */
