/* $Id: exec.c,v 1.15 2002/10/05 04:22:40 cher Exp $ */

/* Copyright (C) 1999-2002 Alexander Chernov <cher@ispras.ru> */
/* Created: <1999-07-20 11:05:09 cher> */

/*
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 */

/**
 * FILE:    reuse/exec-win32.c
 * PURPOSE: process abstraction layer, Win32 implementation
 */

static char const cvsid[]=
"$Id: exec.c,v 1.15 2002/10/05 04:22:40 cher Exp $";

#include "lconfig.h"
#include <p_config.h>
#include CONF_DEPFILE("reuse_version.h")

#include "reuse/exec.h"
#include "reuse/logger.h"
#include "reuse/xalloc.h"
#include "reuse/osdeps.h"
#include "reuse/format_io.h"

#include <windows.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <io.h>

#define NOT_IMPLEMENTED() SWERR(("Not implemented"))

typedef struct tRedir
{
  int fd;			/* file descriptor (0 - 2) */
  int tag;			/* operation */
  union
  {
    int fd2;			/* file descr. (0 - 2) or handle */
    struct
    {
      char *path;		/* path ("/dev/null" included) */
      int   oflag;		/* flags for open(2) */
      int   mode;		/* mode for open(2) (ignored) */
    } s;
    struct
    {
      int    idx;		/* inherited index (0 - read, 1 - write) */
      HANDLE pipe[2];
    } p;
  } u;
} tRedir;
struct redir_arr_s
{
  int     a, u;
  tRedir *v;
};

struct task_args_s
{
  int    a, u;
  char **v;
};

struct tTask
{
  int                 (*main)(int, char **);
  char                 *path;
  int                   state;
  unsigned int          code;
  int                   is_exited; /* is process exited normally */
  int                   exit_code; /* process exit code */
  char                 *cmdline;
  char                 *working_dir;
  PROCESS_INFORMATION   pi;	/* as returned by CreateProcess */
  STARTUPINFO           si;	/* for CreateProcess */
  struct task_args_s    args;
  struct redir_arr_s    redirs;
  HANDLE               *std_streams[3];
  HANDLE                prc;	/* this process handle */
};

struct task_arr_t
{
  int     a, u;
  tTask **v;
};
static struct task_arr_t tasks;

static int               verbose_flag;

#define PROC_SIGNALED(x)   (((x) & 0xC0000000) == 0xC0000000)
#define PROC_SIGNALCODE(x) (((x) & 0x0000FFFF))
#define PROC_EXITED(x)     (((x) & 0xC0000000) != 0xC0000000)
#define PROC_EXITCODE(x)   (((x) & 0x0000FFFF))

  int
task_SetFlag(char *opt, int flag)
{
  switch (flag)
    {
    case 200:
      verbose_flag = 1;
      break;
    case 201:
      verbose_flag = 0;
      break;
    default:
      SWERR(("task_SetFlags: unsupported flag = %d", flag));
    }
  return 0;
}

  tTask *
task_New(void)
{
  tTask *r;

  XCALLOC(r, 1);
  r->state = TSK_STOPPED;

  XEXPAND2(tasks);
  tasks.v[tasks.u++] = r;

  r->std_streams[0] = &r->si.hStdInput;
  r->std_streams[1] = &r->si.hStdOutput;
  r->std_streams[2] = &r->si.hStdError;
  r->prc = GetCurrentProcess();

  return r;
}

  void
task_Delete(tTask *tsk)
{
  int i;

  ASSERT(tsk);

  for (i = 0; i < tasks.u; i++) {
    if (tsk == tasks.v[i]) break;
  }
  if (i >= tasks.u) {
    write_log(LOG_REUSE, LOG_ERROR, "bad task descriptor");
    return;
  }

  /* close all handles */
  for (i = 0; i < tsk->redirs.u; i++) {
    tRedir *p = &tsk->redirs.v[i];
    if (p->tag == TSR_PIPE) {
      if (p->u.p.pipe[0]) CloseHandle(p->u.p.pipe[0]);
      if (p->u.p.pipe[1]) CloseHandle(p->u.p.pipe[1]);
      p->u.p.pipe[0] = 0;
      p->u.p.pipe[1] = 0;
    }
  }
  for (i = 0; i < 3; i++) {
    if (*tsk->std_streams[i]) CloseHandle(*tsk->std_streams[i]);
    *tsk->std_streams[i] = 0;
  }
  if (tsk->prc) CloseHandle(tsk->prc);
  tsk->prc = 0;

  xfree(tsk->path);
  xfree(tsk->cmdline);
  for (i = 0; i < tsk->args.u; i++) {
    xfree(tsk->args.v[i]);
  }
  xfree(tsk->args.v);
  for (i = 0; i < tsk->redirs.u; i++) {
    tRedir *p = &tsk->redirs.v[i];
    switch (p->tag) {
    case TSR_FILE:
      xfree(p->u.s.path);
      break;
    case TSR_PIPE:
      if (p->u.p.pipe[0] != INVALID_HANDLE_VALUE) CloseHandle(p->u.p.pipe[0]);
      if (p->u.p.pipe[1] != INVALID_HANDLE_VALUE) CloseHandle(p->u.p.pipe[1]);
      break;
    }
  }
  xfree(tsk->redirs.v);
  xfree(tsk);
}

/* FIXME: how to implement? return HANDLE? */
  int
task_GetPipe(tTask *tsk, int fd)
{
  int     i;
  tRedir *p;

  ASSERT(tsk);

  for (i = 0; i < tsk->redirs.u; i++) {
    p = &tsk->redirs.v[i];
    if (p->fd == fd && p->tag == TSR_PIPE) {
      write_log(LOG_REUSE, LOG_DEBUG, "task_GetPipe: handle %u",
		(unsigned int) p->u.p.pipe[1 - p->u.p.idx]);
      return (int) p->u.p.pipe[1 - p->u.p.idx];
    }
  }
  
  return -1;
}

  static void
do_add_arg(tTask *tsk, char const *arg)
{
  XEXPAND2(tsk->args);
  tsk->args.v[tsk->args.u++] = xstrdup(arg);
}

  int
task_AddArg(tTask *tsk, char const *arg)
{
  ASSERT(tsk);
  if (!arg) arg = "";
  if (tsk->state != TSK_STOPPED) return -1;

  do_add_arg(tsk, arg);
  return 0;
}

  int
task_nAddArgs(tTask *tsk, int n, ...)
{
  va_list args;

  ASSERT(tsk);
  if (tsk->state != TSK_STOPPED) return -1;

  va_start(args, n);
  for (; n > 0; n--) {
    do_add_arg(tsk, va_arg(args, char *));
  }
  va_end(args);
  return 0;
}

  int
task_zAddArgs(tTask *tsk, ...)
{
  va_list  args;
  char    *s;

  ASSERT(tsk);
  if (tsk->state != TSK_STOPPED) return -1;

  va_start(args, tsk);
  while ((s = va_arg(args, char *)))
    do_add_arg(tsk, s);
  va_end(args);
  return 0;
}

  int
task_pnAddArgs(tTask *tsk, int n, char **p)
{
  ASSERT(tsk);
  if (tsk->state != TSK_STOPPED) return -1;
  if (!p) return 0;

  for (; n > 0; n--) {
    do_add_arg(tsk, *p++);
  }
  return 0;
}

  int
task_pzAddArgs(tTask *tsk, char **p)
{
  ASSERT(tsk);
  if (tsk->state != TSK_STOPPED) return -1;
  if (!p) return 0;

  while (*p) {
    do_add_arg(tsk, *p++);
  }
  return 0;
}

  int
task_SetPath(tTask *tsk, char const *arg)
{
  ASSERT(tsk);
  ASSERT(arg);
  if (tsk->state != TSK_STOPPED) return -1;

  xfree(tsk->path);
  tsk->path = xstrdup(arg);
  return 0;
}

  char *
task_SetPathAsArg0(tTask *tsk)
{
  ASSERT(tsk);
  if (tsk->state != TSK_STOPPED) return NULL;
  if (tsk->args.u <= 0) return NULL;

  xfree(tsk->path);
  tsk->path = xstrdup(tsk->args.v[0]);
  return tsk->path;
}

  int
task_SetWorkingDir(tTask *tsk, char const *path)
{
  ASSERT(tsk);
  if (tsk->state != TSK_STOPPED) return 0;

  xfree(tsk->working_dir);
  tsk->working_dir = xstrdup(path);
  return 0;
}

int
task_PutEnv(tTask *tsk, char const *var)
{
  NOT_IMPLEMENTED();
}

int
task_ClearEnv(tTask *tsk)
{
  NOT_IMPLEMENTED();
}

int
task_SetMaxTime(tTask *tsk, int secs)
{
  NOT_IMPLEMENTED();
}

int
task_SetMaxRealTime(tTask *tsk, int secs)
{
  NOT_IMPLEMENTED();
}

  int
task_SetRedir(tTask *tsk, int fd, int mode, ...)
{
  va_list  args;
  int      i;
  tRedir  *p;

  ASSERT(tsk);
  if (tsk->state != TSK_STOPPED) return -1;

  XEXPAND2(tsk->redirs);
  i = tsk->redirs.u++;
  p = &tsk->redirs.v[i];

  p->fd  = fd;
  p->tag = mode;

  va_start(args, mode);
  switch (mode) {
  case TSR_DUP:
    p->u.fd2 = va_arg(args, int);
    break;
  case TSR_FILE:
    p->u.s.path  = xstrdup(va_arg(args, char *));
    p->u.s.oflag = va_arg(args, int);
    p->u.s.mode  = va_arg(args, int);
    break;
  case TSR_CLOSE:
    break;
  case TSR_PIPE:
    p->u.p.idx     = !!va_arg(args, int);
    p->u.p.pipe[0] = INVALID_HANDLE_VALUE;
    p->u.p.pipe[1] = INVALID_HANDLE_VALUE;
    break;
  default:
    SWERR(("task_SetRedir: mode == %d", mode));
  }
  va_end(args);

  return 0;
}

  int
task_SetEntryFunction(tTask *tsk, int (*func)(int, char **))
{
  ASSERT(tsk);

  tsk->main = func;
  return 0;
}

int
task_DisableCoreDump(tTask *tsk)
{
  ASSERT(tsk);
  /* Win32 does not have core dumps */
  return 0;
}

int
task_SetDataSize(tTask *tsk, int size)
{
  ASSERT(tsk);
  NOT_IMPLEMENTED();
}

int
task_SetStackSize(tTask *tsk, int size)
{
  ASSERT(tsk);
  NOT_IMPLEMENTED();
}

int
task_SetVMSize(tTask *tsk, int size)
{
  ASSERT(tsk);
  NOT_IMPLEMENTED();
}

int
task_SetKillSignal(tTask *tsk, char const *signame)
{
  ASSERT(tsk);
  NOT_IMPLEMENTED();
}

  static char *
print_as_shell_redir(int oflags)
{
  if ((oflags & O_RDWR))
    return "<>";
  if ((oflags & O_APPEND))
    return ">>";
  if ((oflags & O_WRONLY))
    return ">";
  if ((oflags & O_RDONLY))
    return "<";
  return "?";
}

  int
task_PrintArgs(tTask *tsk)
{
  int i;

  ASSERT(tsk);
  if (tsk->state != TSK_STOPPED) return -1;

  if (!tsk->path && tsk->args.u > 0) {
    tsk->path = xstrdup(tsk->args.v[0]);
  }

  if (verbose_flag) {
    if (tsk->main) {
      fprintf(stderr, "task_Start: 0x%08x(%d):",
	      (unsigned int) tsk->main, tsk->args.u);
    } else {
      fprintf(stderr, "task_Start: execv(%d):", tsk->args.u);
    }

    for (i = 0; i < tsk->args.u; i++) {
      fprintf(stderr, " %s", tsk->args.v[i]?tsk->args.v[i]:"<NULL>");
    }

    for (i = 0; i < tsk->redirs.u; i++) {
      tRedir *p = &tsk->redirs.v[i];
      fprintf(stderr, " %u", p->fd);
      switch (p->tag) {
      case TSR_FILE:
	fprintf(stderr, "%s%s",
		print_as_shell_redir(p->u.s.oflag),
		p->u.s.path);
	break;
      case TSR_CLOSE:
	fprintf(stderr, "-");
	break;
      case TSR_DUP:
	fprintf(stderr, ">&%u", p->u.fd2);
	break;
      case TSR_PIPE:
	fprintf(stderr, "%c|", (p->u.p.idx)?'>':'<');
	break;
      default:
	SWERR(("task_PrintArgs: unsupported redirection: %d", p->tag));
      }
    }

    fprintf(stderr, "\n");
  }

  return 0;
}

static int std_handle_names[3] =
{
  STD_INPUT_HANDLE, STD_OUTPUT_HANDLE, STD_ERROR_HANDLE 
};

  HANDLE
normalize_handle(int h)
{
  switch (h) {
  case 0:  return GetStdHandle(STD_INPUT_HANDLE);
  case 1:  return GetStdHandle(STD_OUTPUT_HANDLE);
  case 2:  return GetStdHandle(STD_ERROR_HANDLE);
  default: return (HANDLE) h;
  }
}

  int
close_std_handle(tTask *tsk, int ind)
{
  HANDLE *p;

  ASSERT(tsk);
  ASSERT(ind >= 0 || ind <= 2);
  p = tsk->std_streams[ind];

  if (*p != NULL && *p != INVALID_HANDLE_VALUE) {
    if (!CloseHandle(*p)) {
      write_log(LOG_REUSE, LOG_ERROR,
		"CloseHandle([%d]) failed: %d", ind, GetLastError());
      tsk->state = TSK_ERROR;
      *p = NULL;
      return -1;
    }
  }
  *p = NULL;
  return 0;
}

  int
dup_std_handle(tTask *tsk, int ind, HANDLE src)
{
  HANDLE *p;

  ASSERT(ind >= 0 || ind <= 2);
  p = tsk->std_streams[ind];
  if (src == NULL || src == INVALID_HANDLE_VALUE) {
    src = GetStdHandle(std_handle_names[ind]);
  }

  if (*p == src) return 0;
  if (close_std_handle(tsk, ind) < 0) return -1;

  if (!DuplicateHandle(tsk->prc, src,
		       tsk->prc, p,
		       0, TRUE, DUPLICATE_SAME_ACCESS)) {
    write_log(LOG_REUSE, LOG_ERR,
	      "DuplicateHandle failed: %d", GetLastError());
    tsk->state = TSK_ERROR;
    return -1;
  }
  return 0;
}

  int
task_Start(tTask *tsk)
{
  int     cmdlen = 0;
  int     i;
  HANDLE  hnd = INVALID_HANDLE_VALUE;

  write_log(LOG_REUSE, LOG_DEBUG, "task_Start: entered");

  ASSERT(tsk);
  if (tsk->state != TSK_STOPPED) {
    write_log(LOG_REUSE, LOG_ERROR, "task_Start: bad state");
    return -1;
  }

  if (verbose_flag) task_PrintArgs(tsk);  

  if (tsk->main) {
    int code;

    /* just call the specified function */
    tsk->state = TSK_RUNNING;
    code = tsk->main(tsk->args.u, tsk->args.v);
    tsk->state = TSK_EXITED;
    tsk->is_exited = 1;
    tsk->exit_code = code;
    return 0;
  }

  for (i = 0; i < tsk->args.u; i++) {
    cmdlen += strlen(tsk->args.v[i]) + 5;
  }
  XCALLOC(tsk->cmdline, cmdlen);
  for (i = 0; i < tsk->args.u; i++) {
    if (i) {
      strcat(tsk->cmdline, " ");
    }
    strcat(tsk->cmdline, tsk->args.v[i]);
  }

  /* do default redirections: duplicate standard handles */

  tsk->si.cb         = sizeof(tsk->si);
  tsk->si.dwFlags    = STARTF_USESTDHANDLES;
  if (dup_std_handle(tsk, 0, NULL) < 0) return -1;
  if (dup_std_handle(tsk, 1, NULL) < 0) return -1;
  if (dup_std_handle(tsk, 2, NULL) < 0) return -1;

  /* process user redirections */
  for (i = 0; i < tsk->redirs.u; i++) {
    tRedir *p = &tsk->redirs.v[i];
    ASSERT(p->fd >= 0 && p->fd <= 2);

    switch (p->tag) {
    case TSR_DUP:
      // FIXME: we should manipulate with the new handle names
      switch (p->u.fd2) {
      case 0: hnd = tsk->si.hStdInput; break;
      case 1: hnd = tsk->si.hStdOutput; break;
      case 2: hnd = tsk->si.hStdError; break;
      default:
        SWERR(("not supported"));
        break;
      }
      if (dup_std_handle(tsk,p->fd,hnd) < 0) return -1;
      break;
    case TSR_CLOSE:
      if (close_std_handle(tsk, p->fd) < 0) return -1;
      break;
    case TSR_FILE:
      {
	int                 rw_mode     = 0;
	int                 open_mode   = 0;
	int                 sh_mode     = 0;
	HANDLE              new_hnd;
	SECURITY_ATTRIBUTES attr;

	/* translate some special names */
	if (!strcmp(p->u.s.path, "/dev/null")) {
	  xfree(p->u.s.path);
	  p->u.s.path = xstrdup("NUL");
	} else if (!strcmp(p->u.s.path, "/dev/stdin")) {
	  if (dup_std_handle(tsk, p->fd, normalize_handle(0)) < 0) return -1;
	  break;
	} else if (!strcmp(p->u.s.path, "/dev/stdout")) {
	  if (dup_std_handle(tsk, p->fd, normalize_handle(1)) < 0) return -1;
	  break;
	} else if (!strcmp(p->u.s.path, "/dev/stderr")) {
	  if (dup_std_handle(tsk, p->fd, normalize_handle(2)) < 0) return -1;
	  break;
	}

	switch (p->u.s.oflag) {
	case TSK_REWRITE:
	  rw_mode   = GENERIC_WRITE;
	  open_mode = CREATE_ALWAYS;
	  sh_mode   = FILE_SHARE_READ | FILE_SHARE_WRITE;
	  break;
	case TSK_WRITE:
	  rw_mode   = GENERIC_WRITE;
	  open_mode = OPEN_EXISTING;
	  sh_mode   = FILE_SHARE_READ | FILE_SHARE_WRITE;
	  break;
	default:
	  SWERR(("unsupported open mode: %d", p->u.s.oflag));
	}

	if (close_std_handle(tsk, p->fd) < 0) return -1;

	attr.nLength              = sizeof(attr);
	attr.lpSecurityDescriptor = NULL;
	attr.bInheritHandle       = TRUE;
	new_hnd = CreateFile(p->u.s.path, rw_mode, sh_mode,
			     &attr, open_mode, 0, NULL);
	if (new_hnd == INVALID_HANDLE_VALUE) {
	  write_log(LOG_REUSE, LOG_ERROR,
		    "CreateFile(\"%s\", ...) failed: %d",
		    p->u.s.path, GetLastError());
	  tsk->state = TSK_ERROR;
	  return -1;
	}
	*tsk->std_streams[p->fd] = new_hnd;
      }
      break;
    case TSR_PIPE:
      {
	SECURITY_ATTRIBUTES attr;

	attr.nLength              = sizeof(attr);
	attr.lpSecurityDescriptor = NULL;
	attr.bInheritHandle       = TRUE;

	close_std_handle(tsk, p->fd);
	if (!(CreatePipe(&p->u.p.pipe[0], &p->u.p.pipe[1], &attr, 0))) {
	  write_log(LOG_REUSE, LOG_ERROR, "CreatePipe failed: %d",
		    GetLastError());
	  tsk->state = TSK_ERROR;
	  return -1;
	}

	/* make non-inheritable the other end of pipe */
	SetHandleInformation(p->u.p.pipe[1 - p->u.p.idx],
			     HANDLE_FLAG_INHERIT, 0);
	*tsk->std_streams[p->fd] = p->u.p.pipe[p->u.p.idx];
      }
      break;
    default:
      SWERR(("task_Start: unsupported redirection: %d", p->tag));
    }
  }

  if (!CreateProcess(NULL, tsk->cmdline, NULL, NULL,
		     TRUE, CREATE_NEW_PROCESS_GROUP,
		     NULL, tsk->working_dir,
                     &tsk->si, &tsk->pi)) {
    write_log(LOG_REUSE, LOG_ERROR, "CreateProcess failed: %d",
	      GetLastError());
    tsk->state = TSK_ERROR;
    tsk->code = GetLastError();
    return -1;
  }

  /* close the opposite side of pipes */
  for (i = 0; i < tsk->redirs.u; i++) {
    tRedir *p = &tsk->redirs.v[i];
    if (p->tag == TSR_PIPE) {
      HANDLE hh = p->u.p.pipe[p->u.p.idx];
      if (hh != NULL && hh != INVALID_HANDLE_VALUE) {
	CloseHandle(hh);
	p->u.p.pipe[p->u.p.idx] = NULL;
      }
    }
  }

  /* close standard descriptors */
  if (tsk->si.hStdInput != INVALID_HANDLE_VALUE) {
    CloseHandle(tsk->si.hStdInput);
    tsk->si.hStdInput = INVALID_HANDLE_VALUE;
  }
  if (tsk->si.hStdOutput != INVALID_HANDLE_VALUE) {
    CloseHandle(tsk->si.hStdOutput);
    tsk->si.hStdOutput = INVALID_HANDLE_VALUE;
  }
  if (tsk->si.hStdError != INVALID_HANDLE_VALUE) {
    CloseHandle(tsk->si.hStdError);
    tsk->si.hStdError = INVALID_HANDLE_VALUE;
  }

  write_log(LOG_REUSE, LOG_INFO, "new process started: %u",
	    (unsigned int) tsk->pi.hProcess);
  tsk->state = TSK_RUNNING;
  return 0;
}

  tTask *
task_Wait(tTask *tsk)
{
  ASSERT(tsk);

  if (tsk->state == TSK_ERROR || tsk->state == TSK_STOPPED) return NULL;
  if (tsk->state == TSK_SIGNALED || tsk->state == TSK_EXITED) return tsk;

  ASSERT(tsk->state == TSK_RUNNING);
  if (WaitForSingleObject(tsk->pi.hProcess, INFINITE) == WAIT_FAILED) {
    tsk->state = TSK_ERROR;
    tsk->code = GetLastError();
    write_log(LOG_REUSE, LOG_ERROR, "WaitForSingleObject failed: %d",
	      GetLastError());
    return NULL;
  }

  GetExitCodeProcess(tsk->pi.hProcess, (DWORD*) &tsk->code);
  if (PROC_SIGNALED(tsk->code)) {
    tsk->state = TSK_SIGNALED;
  } else {
    tsk->state = TSK_EXITED;
  }
  return tsk;
}

  int
task_Status(tTask *tsk)
{
  ASSERT(tsk);
  return tsk->state;
}

  int
task_TermSignal(tTask *tsk)
{
  ASSERT(tsk);
  if (tsk->state != TSK_SIGNALED) return -1;
  return PROC_SIGNALCODE(tsk->code);
}

  int
task_ExitCode(tTask *tsk)
{
  ASSERT(tsk);
  if (tsk->state != TSK_EXITED)
    return -1;
  if (tsk->is_exited)
    return tsk->exit_code;
  return PROC_EXITCODE(tsk->code);
}

int
task_IsTimeout(tTask *tsk)
{
  NOT_IMPLEMENTED();
}

long
task_GetRunningTime(tTask *tsk)
{
  /* GetProcessTimes */
  NOT_IMPLEMENTED();
}

  int
task_IsAbnormal(tTask *tsk)
{
  ASSERT(tsk);
  if (tsk->state == TSK_SIGNALED)
    return 1;
  if (tsk->state == TSK_EXITED && tsk->is_exited && tsk->exit_code > 0)
    return 1;
  if (tsk->state == TSK_EXITED && tsk->code > 0)
    return 1;
  return 0;
}

  void
task_Log(tTask *tsk, int fac, int sev)
{
  ASSERT(tsk);

  if (tsk->state != TSK_SIGNALED && tsk->state != TSK_EXITED)
    return;

  if (tsk->state == TSK_SIGNALED) {
      write_log(fac, sev,
                "process %u is terminated with signal %d (%s)",
                (unsigned int) tsk->pi.hProcess,
                PROC_SIGNALCODE(tsk->code),
                os_GetSignalString(PROC_SIGNALCODE(tsk->code)));
  } else {
      int   r = tsk->code;
      if (tsk->is_exited) r = tsk->exit_code;
      write_log(fac, sev,
                "process %u is exited with code %d",
                (unsigned int) tsk->pi.hProcess, PROC_EXITCODE(r));
    }
}

int
task_ErrorCode(tTask *tsk, int *p_exit_code, int *p_error)
{
  NOT_IMPLEMENTED();
}

/* === Do not add stuff after this line === */

/*
 * Local variables:
 *  compile-command: "make -C .."
 *  c-font-lock-extra-types: ("HANDLE" "FILE" "\\sw+_t" "tRedir" "tTask" "PROCESS_INFORMATION" "STARTUPINFO" "SECURITY_ATTRIBUTES" "DWORD")
 * End:
 */
